# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Optional
from datetime import datetime
from dataclasses import dataclass
from warnings import warn

from matplotlib.colors import LogNorm
import numpy as np

from spectre_core.spectrograms import Spectrogram, FrequencyCut, TimeCut
from ._base import BasePanel, BaseSpectrumPanel, BaseTimeSeriesPanel


@dataclass(frozen=True)
class PanelNames:
    SPECTROGRAM            : str = "spectrogram"
    FREQUENCY_CUTS         : str = "frequency_cuts"
    TIME_CUTS              : str = "time_cuts"
    INTEGRAL_OVER_FREQUENCY: str = "integral_over_frequency"


class _FrequencyCutsPanel(BaseSpectrumPanel):
    def __init__(self, 
                 spectrogram: Spectrogram, 
                 *times: list[float | str],
                 dBb: bool = False,
                 peak_normalise: bool = False):
        super().__init__(PanelNames.FREQUENCY_CUTS,
                         spectrogram)
        self._times = times
        self._dBb = dBb
        self._peak_normalise = peak_normalise
        # map each time cut to the corresponding FrequencyCut dataclass
        self._frequency_cuts: Optional[dict[float | datetime, FrequencyCut]] = {}


    @property
    def frequency_cuts(self) -> dict[float | str, FrequencyCut]:
        if not self._frequency_cuts:
            for time in self._times:
                frequency_cut = self._spectrogram.get_frequency_cut(time,
                                                                    dBb = self._dBb,
                                                                    peak_normalise = self._peak_normalise)
                self._frequency_cuts[frequency_cut.time] = frequency_cut
        return self._frequency_cuts


    @property
    def times(self) -> list[float | datetime]:
        return list(self.frequency_cuts.keys())
    

    def draw(self):
        for time, color in self.bind_to_colors():
            frequency_cut = self.frequency_cuts[time]
            self.ax.step(self.frequencies*1e-6, # convert to MHz
                         frequency_cut.cut, 
                         where='mid', 
                         color = color)
    

    def annotate_y_axis(self) -> None:
        if self._dBb:
            self.ax.set_ylabel('dBb')
        elif self._peak_normalise:
            return # no y-axis label
        else:
            self.ax.set_ylabel(f'{self._spectrogram.spectrum_type.capitalize()}')

    
    def bind_to_colors(self):
        return super().bind_to_colors(self.times, cmap = self.panel_format.cuts_cmap)
    

class _IntegralOverFrequencyPanel(BaseTimeSeriesPanel):
    def __init__(self, 
                 spectrogram: Spectrogram, 
                 peak_normalise: bool = False,
                 background_subtract: bool = False):
        super().__init__(PanelNames.INTEGRAL_OVER_FREQUENCY,
                         spectrogram)
        self._peak_normalise = peak_normalise
        self._background_subtract = background_subtract


    def draw(self):
        I = self._spectrogram.integrate_over_frequency(correct_background = self._background_subtract,
                                                       peak_normalise = self._peak_normalise)
        self.ax.step(self.times, I, where="mid", color = self.panel_format.integral_color)
 

    def annotate_y_axis(self):
        # no y-axis label
        return 


class _TimeCutsPanel(BaseTimeSeriesPanel):
    def __init__(self, 
                 spectrogram: Spectrogram, 
                 *frequencies: list[float],
                 dBb: bool = False,
                 peak_normalise: bool = False,
                 background_subtract: bool = False):
        super().__init__(PanelNames.TIME_CUTS, 
                         spectrogram)
        self._frequencies = frequencies
        self._dBb = dBb
        self._peak_normalise = peak_normalise
        self._background_subtract = background_subtract
        # map each cut frequency to the corresponding TimeCut dataclass
        self._time_cuts: Optional[dict[float, TimeCut]] = {} 
    

    @property
    def time_cuts(self) -> dict[float, TimeCut]:
        if not self._time_cuts:
            for frequency in self._frequencies:
                time_cut =  self._spectrogram.get_time_cut(frequency,
                                                           dBb = self._dBb,
                                                           peak_normalise = self._peak_normalise,
                                                           correct_background = self._background_subtract,
                                                           return_time_type=self._time_type)
                self._time_cuts[time_cut.frequency] = time_cut
        return self._time_cuts
    

    @property
    def frequencies(self) -> list[float]:
        return list(self.time_cuts.keys())


    def draw(self):
        for frequency, color in self.bind_to_colors():
            time_cut = self.time_cuts[frequency]
            self.ax.step(self.times, 
                         time_cut.cut, 
                         where='mid', 
                         color = color)
    

    def annotate_y_axis(self) -> None:
        if self._dBb:
            self.ax.set_ylabel('dBb')
        elif self._peak_normalise:
            return # no y-axis label
        else:
            self.ax.set_ylabel(f'{self._spectrogram.spectrum_type.capitalize()}')

    
    def bind_to_colors(self):
        return super().bind_to_colors(self.frequencies, cmap = self.panel_format.cuts_cmap)
    

class _SpectrogramPanel(BaseTimeSeriesPanel):
    def __init__(self, 
                 spectrogram: Spectrogram, 
                 log_norm: bool = False,
                 dBb: bool = False,
                 vmin: float | None = -1,
                 vmax: float | None = 2):
        super().__init__(PanelNames.SPECTROGRAM,
                         spectrogram)
        self._log_norm = log_norm
        self._dBb = dBb
        self._vmin = vmin
        self._vmax = vmax


    def draw(self):
        dynamic_spectra = self._spectrogram.dynamic_spectra_as_dBb if self._dBb else self._spectrogram.dynamic_spectra

        norm = LogNorm(vmin=np.nanmin(dynamic_spectra[dynamic_spectra > 0]), 
                       vmax=np.nanmax(dynamic_spectra)) if self._log_norm else None
        

        if self._log_norm and (self._vmin or self._vmax):
            warn("vmin/vmax will be ignored while using log_norm.")
            self._vmin = None 
            self._vmax = None
        
        # Plot the spectrogram with kwargs
        pcm = self.ax.pcolormesh(self.times, 
                            self._spectrogram.frequencies * 1e-6, 
                            dynamic_spectra,
                            vmin=self._vmin, 
                            vmax=self._vmax,
                            norm=norm, 
                            cmap=self.panel_format.spectrogram_cmap)
        
        # Add colorbar if dBb is used
        if self._dBb:
            cbar = self.fig.colorbar(pcm, 
                                      ax=self.ax, 
                                      ticks=np.linspace(self._vmin, self._vmax, 6, dtype=int))
            cbar.set_label('dBb')


    def annotate_y_axis(self) -> None:
        self.ax.set_ylabel('Frequency [MHz]')
        return
    
    
    def overlay_cuts(self, cuts_panel: BasePanel) -> None:
        if cuts_panel.name == PanelNames.TIME_CUTS:
            self._overlay_time_cuts(cuts_panel)
        elif cuts_panel.name == PanelNames.FREQUENCY_CUTS:
            self._overlay_frequency_cuts(cuts_panel)


    def _overlay_time_cuts(self, time_cuts_panel: _TimeCutsPanel) -> None:
        for frequency, color in time_cuts_panel.bind_to_colors():
            self.ax.axhline(frequency*1e-6, # convert to MHz
                            color = color,
                            linewidth=self.panel_format.line_width
                            )
            
            
    def _overlay_frequency_cuts(self, frequency_cuts_panel: _FrequencyCutsPanel) -> None:
        for time, color in frequency_cuts_panel.bind_to_colors():
            self.ax.axvline(time,
                            color = color,
                            linewidth=self.panel_format.line_width
                            )


@dataclass(frozen=True)
class Panels:
    TimeCuts                    = _TimeCutsPanel
    FrequencyCuts               = _FrequencyCutsPanel
    Spectrogram                 = _SpectrogramPanel
    IntegralOverFrequency       = _IntegralOverFrequencyPanel


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Internal file handling.
"""

from .file_handlers import (
    BaseFileHandler, JsonHandler, TextHandler,
)

__all__ = [ 
    "BaseFileHandler", "JsonHandler", "TextHandler"
]


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Callable
from dataclasses import dataclass

import numpy as np

from spectre_core.capture_configs import CaptureConfig, PNames
from spectre_core.exceptions import ModeNotFoundError
from ._spectrogram import Spectrogram
from ._array_operations import is_close

__all__ = [
    "get_analytical_spectrogram",
    "validate_analytically",
    "TestResults"
]

@dataclass
class TestResults:
    # Whether the times array matches analytically 
    times_validated: bool = False  
    # Whether the frequencies array matches analytically
    frequencies_validated: bool = False  
    # Maps each time to whether the corresponding spectrum matched analytically
    spectrum_validated: dict[float, bool] = None


    @property
    def num_validated_spectrums(self) -> int:
        """Counts the number of validated spectrums."""
        return sum(is_validated for is_validated in self.spectrum_validated.values())


    @property
    def num_invalid_spectrums(self) -> int:
        """Counts the number of spectrums that are not validated."""
        return len(self.spectrum_validated) - self.num_validated_spectrums
    

    def to_dict(self) -> dict[str, bool | dict[float, bool]]:
        return {
            "times_validated": self.times_validated,
            "frequencies_validated": self.frequencies_validated,
            "spectrum_validated": self.spectrum_validated
        }


class _AnalyticalFactory:
    def __init__(self):
        self._builders: dict[str, Callable] = {
            "cosine-signal-1": self.cosine_signal_1,
            "tagged-staircase": self.tagged_staircase
        }
        self._test_modes = list(self.builders.keys())


    @property
    def builders(self) -> dict[str, Callable]:
        return self._builders
    

    @property
    def test_modes(self) -> list[str]:
        return self._test_modes
    

    def get_spectrogram(self, 
                        num_spectrums: int, 
                        capture_config: CaptureConfig) -> Spectrogram:
        """Get an analytical spectrogram based on a test receiver capture config.
        
        The anaytically derived spectrogram should be able to be fully determined
        by parameters in the corresponding capture config and the number of spectrums
        in the output spectrogram.
        """

        if capture_config.receiver_name != "test":
            raise ValueError(f"Input capture config must correspond to the test receiver")
        
        builder_method = self.builders.get(capture_config.receiver_mode)
        if builder_method is None:
            raise ModeNotFoundError(f"Test mode not found. Expected one of {self.test_modes}, but received {capture_config.receiver_mode}")
        return builder_method(num_spectrums, 
                              capture_config)
    

    def cosine_signal_1(self, 
                        num_spectrums: int,
                        capture_config: CaptureConfig) -> Spectrogram:
        # Extract necessary parameters from the capture configuration.
        window_size      = capture_config.get_parameter_value(PNames.WINDOW_SIZE)
        sample_rate      = capture_config.get_parameter_value(PNames.SAMPLE_RATE)
        amplitude        = capture_config.get_parameter_value(PNames.AMPLITUDE)
        frequency        = capture_config.get_parameter_value(PNames.FREQUENCY)
        window_hop       = capture_config.get_parameter_value(PNames.WINDOW_HOP)
        center_frequency = capture_config.get_parameter_value(PNames.CENTER_FREQUENCY)
        # Calculate derived parameters a (sampling rate ratio) and p (sampled periods).
        a = int(sample_rate / frequency)
        p = int(window_size / a)

        # Create the analytical spectrum, which is constant in time.
        spectrum = np.zeros(window_size)
        spectral_amplitude = amplitude * window_size / 2
        spectrum[p] = spectral_amplitude
        spectrum[window_size - p] = spectral_amplitude

        # Align spectrum to naturally ordered frequency array.
        spectrum = np.fft.fftshift(spectrum)

        # Populate the spectrogram with identical spectra.
        analytical_dynamic_spectra = np.ones((window_size, num_spectrums)) * spectrum[:, np.newaxis]

        # Compute time array.
        sampling_interval = 1 / sample_rate
        times = np.arange(num_spectrums) * window_hop * sampling_interval

        # compute the frequency array.
        frequencies = np.fft.fftshift(np.fft.fftfreq(window_size, sampling_interval)) + center_frequency

        # Return the spectrogram.
        return Spectrogram(analytical_dynamic_spectra,
                           times,
                           frequencies,
                           'analytically-derived-spectrogram',
                           spectrum_type="amplitude")


    def tagged_staircase(self, 
                        num_spectrums: int,
                        capture_config: CaptureConfig) -> Spectrogram:
        # Extract necessary parameters from the capture configuration.
        window_size          = capture_config.get_parameter_value(PNames.WINDOW_SIZE)
        min_samples_per_step = capture_config.get_parameter_value(PNames.MIN_SAMPLES_PER_STEP)
        max_samples_per_step = capture_config.get_parameter_value(PNames.MAX_SAMPLES_PER_STEP)
        step_increment       = capture_config.get_parameter_value(PNames.STEP_INCREMENT)
        samp_rate            = capture_config.get_parameter_value(PNames.SAMPLE_RATE)

        # Calculate step sizes and derived parameters.
        num_samples_per_step = np.arange(min_samples_per_step, max_samples_per_step + 1, step_increment)
        num_steps = len(num_samples_per_step)

        # Create the analytical spectrum, constant in time.
        spectrum = np.zeros(window_size * num_steps)
        step_count = 0
        for i in range(num_steps):
            step_count += 1
            spectral_amplitude = window_size * step_count
            spectrum[int(window_size/2) + i*window_size] = spectral_amplitude

        # Populate the spectrogram with identical spectra.
        analytical_dynamic_spectra = np.ones((window_size * num_steps, num_spectrums)) * spectrum[:, np.newaxis]

        # Compute time array
        num_samples_per_sweep = sum(num_samples_per_step)
        sampling_interval = 1 / samp_rate
        # compute the sample index we are "assigning" to each spectrum
        # and multiply by the sampling interval to get the equivalent physical time
        times = np.array([(i * num_samples_per_sweep) for i in range(num_spectrums) ]) * sampling_interval

        # Compute the frequency array
        baseband_frequencies = np.fft.fftshift(np.fft.fftfreq(window_size, sampling_interval))
        frequencies = np.empty((window_size * num_steps))
        for i in range(num_steps):
            lower_bound = i * window_size
            upper_bound = (i + 1) * window_size
            frequencies[lower_bound:upper_bound] = baseband_frequencies + (samp_rate / 2) + (samp_rate * i)

        # Return the spectrogram.
        return Spectrogram(analytical_dynamic_spectra,
                           times,
                           frequencies,
                           'analytically-derived-spectrogram',
                           spectrum_type="amplitude")
    

def get_analytical_spectrogram(num_spectrums: int,
                               capture_config: CaptureConfig) -> Spectrogram:
    
    factory = _AnalyticalFactory()
    return factory.get_spectrogram(num_spectrums,
                                   capture_config)


def validate_analytically(spectrogram: Spectrogram,
                          capture_config: CaptureConfig,
                          absolute_tolerance: float) -> TestResults:

    analytical_spectrogram = get_analytical_spectrogram(spectrogram.num_times,
                                                        capture_config)


    test_results = TestResults()

    test_results.times_validated = bool(is_close(analytical_spectrogram.times,
                                                 spectrogram.times,
                                                 absolute_tolerance))

    test_results.frequencies_validated = bool(is_close(analytical_spectrogram.frequencies,
                                                       spectrogram.frequencies,
                                                       absolute_tolerance))

    test_results.spectrum_validated = {}
    for i in range(spectrogram.num_times):
        time = spectrogram.times[i]
        analytical_spectrum = analytical_spectrogram.dynamic_spectra[:, i]
        spectrum = spectrogram.dynamic_spectra[:, i]
        test_results.spectrum_validated[time] = bool(is_close(analytical_spectrum, 
                                                              spectrum,
                                                              absolute_tolerance))

    return test_results


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

# register decorators take effect on import
from .library._test import _Receiver
from .library._rsp1a import _Receiver
from .library._rspduo import _Receiver

from ._base import BaseReceiver
from ._factory import get_receiver
from ._register import list_all_receiver_names
from ._spec_names import SpecNames

__all__ = [
    "BaseReceiver", "get_receiver", "list_all_receiver_names", "SpecNames"
]


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
from datetime import datetime, timedelta
from typing import Optional
from math import floor

from spectre_core.config import TimeFormats
from ._array_operations import find_closest_index, average_array
from ._spectrogram import Spectrogram

__all__ = [
    "frequency_chop",
    "time_chop",
    "frequency_average",
    "time_average",
    "join_spectrograms"
]

def frequency_chop(input_spectrogram: Spectrogram, 
                   start_frequency: float | int, 
                   end_frequency: float | int) -> Optional[Spectrogram]:
    
    is_entirely_below_frequency_range = (start_frequency < input_spectrogram.frequencies[0] and end_frequency < input_spectrogram.frequencies[0])
    is_entirely_above_frequency_range = (start_frequency > input_spectrogram.frequencies[-1] and end_frequency > input_spectrogram.frequencies[-1])
    # if the requested frequency range is out of bounds for the spectrogram return None
    if is_entirely_below_frequency_range or is_entirely_above_frequency_range:
        return None
    
    #find the index of the nearest matching frequency bins in the spectrogram
    start_index = find_closest_index(start_frequency, input_spectrogram.frequencies)
    end_index = find_closest_index(end_frequency, input_spectrogram.frequencies)
    
    # enforce distinct start and end indices
    if start_index == end_index:
        raise ValueError(f"Start and end indices are equal! Got start_index: {start_index} and end_index: {end_index}")  
    
    # if start index is more than end index, swap the ordering so to enforce start_index <= end_index
    if start_index > end_index:
        start_index, end_index = end_index, start_index
    
    # chop the spectrogram accordingly
    transformed_dynamic_spectra = input_spectrogram.dynamic_spectra[start_index:end_index+1, :]
    transformed_frequencies = input_spectrogram.frequencies[start_index:end_index+1]
    
    # return the spectrogram instance
    return Spectrogram(transformed_dynamic_spectra,
                       input_spectrogram.times,
                       transformed_frequencies,
                       input_spectrogram.tag,
                       chunk_start_time = input_spectrogram.chunk_start_time,
                       microsecond_correction = input_spectrogram.microsecond_correction,
                       spectrum_type = input_spectrogram.spectrum_type)


def time_chop(input_spectrogram: Spectrogram, 
              start_time: str, 
              end_time: str, 
              time_format: str = TimeFormats.DATETIME) -> Optional[Spectrogram]:
    
    # parse the strings as datetimes
    start_datetime = datetime.strptime(start_time, time_format)
    end_datetime = datetime.strptime(end_time, time_format)

    # if the requested time range is out of bounds for the spectrogram return None
    is_entirely_below_time_range = (start_datetime < input_spectrogram.datetimes[0] and end_datetime < input_spectrogram.datetimes[0])
    is_entirely_above_time_range = (start_datetime > input_spectrogram.datetimes[-1] and end_datetime > input_spectrogram.datetimes[-1])
    if is_entirely_below_time_range or is_entirely_above_time_range:
        return None
    
    start_index = find_closest_index(start_datetime, input_spectrogram.datetimes)
    end_index = find_closest_index(end_datetime, input_spectrogram.datetimes)
    
    if start_index == end_index:
        raise ValueError(f"Start and end indices are equal! Got start_index: {start_index} and end_index: {end_index}")
    
    if start_index > end_index:
        start_index, end_index = end_index, start_index

    # chop the spectrogram 
    transformed_dynamic_spectra = input_spectrogram.dynamic_spectra[:, start_index:end_index+1]

    # compute the new start datetime following the time chop
    transformed_start_datetime = input_spectrogram.datetimes[start_index]
    # compute the microsecond correction, and chunk start time
    transformed_chunk_start_time = datetime.strftime(transformed_start_datetime, TimeFormats.DATETIME)
    transformed_microsecond_correction = transformed_start_datetime.microsecond

    # chop the times array
    transformed_times = input_spectrogram.times[start_index:end_index+1]
    # assign the first spectrum to t=0 [s]
    transformed_times -= transformed_times[0]

    return Spectrogram(transformed_dynamic_spectra, 
                       transformed_times, 
                       input_spectrogram.frequencies, 
                       input_spectrogram.tag, 
                       chunk_start_time = transformed_chunk_start_time,
                       microsecond_correction = transformed_microsecond_correction,
                       spectrum_type = input_spectrogram.spectrum_type)


def time_average(input_spectrogram: Spectrogram, 
                 resolution: Optional[float] = None,
                 average_over: Optional[int] = None) -> Spectrogram:

    # spectre does not currently support averaging of non-datetime assigned spectrograms
    if input_spectrogram.chunk_start_time is None:
        raise ValueError(f"Input spectrogram is missing chunk start time. Averaging is not yet supported for non-datetime assigned spectrograms")
    
    # if nothing is specified, do nothing
    if (resolution is None) and (average_over is None):
        average_over = 1

    if not (resolution is not None) ^ (average_over is not None):
        raise ValueError(f"Exactly one of 'resolution' or 'average_over' "
                         f"must be specified.")
    
    # if the resolution is specified, compute the appropriate number of spectrums to average over
    # and recall the same function
    if resolution is not None:
        average_over = max(1, floor(resolution / input_spectrogram.time_resolution))
        return time_average(input_spectrogram, average_over=average_over)
    
    # No averaging is required, if we have to average over every one spectrum
    if average_over == 1:
        return input_spectrogram

    # average the dynamic spectra array
    transformed_dynamic_spectra = average_array(input_spectrogram.dynamic_spectra, 
                                                 average_over, 
                                                 axis=1)

    # We need to assign timestamps to the averaged spectrums in the spectrograms. 
    # The natural way to do this is to assign the i'th averaged spectrogram 
    # to the i'th averaged time
    transformed_times = average_array(input_spectrogram.times, average_over)
    
    # find the new chunk start time, which we will assign to the first spectrum after averaging
    corrected_start_datetime = input_spectrogram.datetimes[0] + timedelta(seconds = float(transformed_times[0]))
    transformed_chunk_start_time = corrected_start_datetime.strftime(TimeFormats.DATETIME)
    transformed_microsecond_correction = corrected_start_datetime.microsecond

    # finally, translate the averaged time seconds to begin at t=0 [s]
    transformed_times -= transformed_times[0]
    return Spectrogram(transformed_dynamic_spectra, 
                       transformed_times, 
                       input_spectrogram.frequencies, 
                       input_spectrogram.tag,
                       chunk_start_time = transformed_chunk_start_time,
                       microsecond_correction = transformed_microsecond_correction,
                       spectrum_type = input_spectrogram.spectrum_type)



def frequency_average(input_spectrogram: Spectrogram, 
                      resolution: Optional[float] = None,
                      average_over: Optional[int] = None) -> Spectrogram:
    
    # if nothing is specified, do nothing
    if (resolution is None) and (average_over is None):
        average_over = 1

    if not (resolution is not None) ^ (average_over is not None):
        raise ValueError(f"Exactly one of 'resolution' or 'average_over' "
                         f"must be specified.")
    
    # if the resolution is specified, compute the appropriate number of spectrums to average over
    # and recall the same function
    if resolution is not None:
        average_over = max(1, floor(resolution / input_spectrogram.frequency_resolution))
        return frequency_average(input_spectrogram, average_over=average_over)
    
    # No averaging is required, if we have to average over every one spectrum
    if average_over == 1:
        return input_spectrogram
    
    # We need to assign physical frequencies to the averaged spectrums in the spectrograms.
    # is to assign the i'th averaged spectral component to the i'th averaged frequency.
    # average the dynamic spectra array
    transformed_dynamic_spectra = average_array(input_spectrogram.dynamic_spectra, 
                                                 average_over, 
                                                 axis=0)
    transformed_frequencies = average_array(input_spectrogram.frequencies, average_over)

    return Spectrogram(transformed_dynamic_spectra, 
                       input_spectrogram.times, 
                       transformed_frequencies, 
                       input_spectrogram.tag,
                       chunk_start_time = input_spectrogram.chunk_start_time, 
                       microsecond_correction = input_spectrogram.microsecond_correction,
                       spectrum_type = input_spectrogram.spectrum_type)


def _time_elapsed(datetimes: np.ndarray) -> np.ndarray:
    # Extract the first datetime to use as the reference point
    base_time = datetimes[0]
    # Calculate elapsed time in seconds for each datetime in the list
    elapsed_time = [(dt - base_time).total_seconds() for dt in datetimes]
    # Convert the list of seconds to a NumPy array of type float32
    return np.array(elapsed_time, dtype=np.float32)


# we assume that the spectrogram list is ORDERED chronologically
# we assume there is no time overlap in any of the spectrograms in the list
def join_spectrograms(spectrograms: list[Spectrogram]) -> Spectrogram:

    # check that the length of the list is non-zero
    num_spectrograms = len(spectrograms)
    if num_spectrograms == 0:
        raise ValueError(f"Input list of spectrograms is empty!")
    
    # extract the first element of the list, and use this as a reference for comparison
    # input validations.
    reference_spectrogram = spectrograms[0] 

    # perform checks on each spectrogram in teh list
    for spectrogram in spectrograms:
        if not np.all(np.equal(spectrogram.frequencies, reference_spectrogram.frequencies)):
            raise ValueError(f"All spectrograms must have identical frequency ranges")
        if spectrogram.tag != reference_spectrogram.tag:
            raise ValueError(f"All tags must be equal for each spectrogram in the input list!")
        if spectrogram.spectrum_type != reference_spectrogram.spectrum_type:
            raise ValueError(f"All units must be equal for each spectrogram in the input list!")
        if spectrogram.chunk_start_time is None:
            raise ValueError(f"All spectrograms must have chunk_start_time set. Received {spectrogram.chunk_start_time}")


    # build a list of the time array of each spectrogram in the list
    conc_datetimes = np.concatenate([s.datetimes for s in spectrograms])
    # find the total number of time stamps
    num_total_time_bins = len(conc_datetimes)
    # find the total number of frequency bins (we can safely now use the first)
    num_total_freq_bins = len(reference_spectrogram.frequencies)
    # create an empty numpy array to hold the joined spectrograms
    transformed_dynamic_spectra = np.empty((num_total_freq_bins, num_total_time_bins))

    start_index = 0
    for spectrogram in spectrograms:
        end_index = start_index + len(spectrogram.times)
        transformed_dynamic_spectra[:, start_index:end_index] = spectrogram.dynamic_spectra
        start_index = end_index

    transformed_times = _time_elapsed(conc_datetimes)

    return Spectrogram(transformed_dynamic_spectra, 
                       transformed_times, 
                       reference_spectrogram.frequencies, 
                       reference_spectrogram.tag, 
                       chunk_start_time = reference_spectrogram.chunk_start_time,
                       microsecond_correction = reference_spectrogram.microsecond_correction,
                       spectrum_type = reference_spectrogram.spectrum_type) 


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Optional
from warnings import warn
from datetime import datetime, timedelta
from dataclasses import dataclass
import os

import numpy as np
from astropy.io import fits

from spectre_core.capture_configs import CaptureConfig, PNames
from spectre_core.config import get_chunks_dir_path, TimeFormats
from ._array_operations import (
    find_closest_index,
    normalise_peak_intensity,
    compute_resolution,
    compute_range,
    subtract_background,
)

__all__ = [
    "FrequencyCut",
    "TimeCut",
    "TimeTypes",
    "SpectrumTypes",
    "Spectrogram"
]

@dataclass
class FrequencyCut:
    time: float | datetime
    frequencies: np.ndarray
    cut: np.ndarray
    spectrum_type: str


@dataclass
class TimeCut:
    frequency: float
    times: np.ndarray
    cut: np.ndarray
    spectrum_type: str


@dataclass(frozen=True)
class TimeTypes:
    SECONDS  : str = "seconds"
    DATETIMES: str = "datetimes"
    

@dataclass(frozen=True)
class SpectrumTypes:
    AMPLITUDE: str = "amplitude"
    POWER    : str = "power"
    DIGITS   : str = "digits"


class Spectrogram:
    def __init__(self, 
                 dynamic_spectra: np.ndarray, # holds the spectrogram data
                 times: np.ndarray, # holds the time stamp [s] for each spectrum
                 frequencies: np.ndarray,  # physical frequencies [Hz] for each spectral component
                 tag: str,
                 chunk_start_time: Optional[str] = None, 
                 microsecond_correction: int = 0, 
                 spectrum_type: Optional[str] = None,
                 start_background: Optional[str] = None, 
                 end_background: Optional[str] = None): 
        
        # dynamic spectra
        self._dynamic_spectra = dynamic_spectra
        self._dynamic_spectra_as_dBb: Optional[np.ndarray] = None # cache

        # assigned times and frequencies
        if times[0] != 0:
            raise ValueError(f"The first spectrum must correspond to t=0 [s]")
        
        self._times = times
        self._datetimes: Optional[list[datetime]] = None # cache
        self._frequencies = frequencies

        # general metadata
        self._tag = tag
        self._chunk_start_time = chunk_start_time
        self._chunk_start_datetime: Optional[datetime] = None # cache
        self._microsecond_correction = microsecond_correction
        self._spectrum_type = spectrum_type

        # background metadata     
        self._background_spectrum: Optional[np.ndarray] = None # cache
        self._start_background = start_background
        self._end_background = end_background
        self._start_background_index = 0 # by default
        self._end_background_index = self.num_times # by default
        self._check_shapes()


    @property
    def dynamic_spectra(self) -> np.ndarray:
        return self._dynamic_spectra
    

    @property
    def times(self) -> np.ndarray:
        return self._times
    
    
    @property
    def num_times(self) -> int:
        return len(self._times)
    

    @property
    def time_resolution(self) -> float:
        return compute_resolution(self._times)
    

    @property
    def time_range(self) -> float:
        return compute_range(self._times)
    

    @property
    def datetimes(self) -> list[datetime]:
        if self._datetimes is None:
            self._datetimes = [self.chunk_start_datetime + timedelta(seconds=(t + self.microsecond_correction*1e-6)) for t in self._times]
        return self._datetimes
    

    @property
    def frequencies(self) -> np.ndarray:
        return self._frequencies


    @property
    def num_frequencies(self) -> int:
        return len(self._frequencies)
    
    
    @property
    def frequency_resolution(self) -> float:
        return compute_resolution(self._frequencies)
    

    @property
    def frequency_range(self) -> float:
        return compute_range(self._frequencies)
    

    @property
    def tag(self) -> str:
        return self._tag
    

    @property
    def chunk_start_time(self) -> str:
        if self._chunk_start_time is None:
            raise AttributeError(f"Chunk start time has not been set.")
        return self._chunk_start_time
    

    @property
    def chunk_start_datetime(self) -> datetime:
        if self._chunk_start_datetime is None:
            self._chunk_start_datetime = datetime.strptime(self.chunk_start_time, TimeFormats.DATETIME)
        return self._chunk_start_datetime


    @property
    def microsecond_correction(self) -> int:
        return self._microsecond_correction
    

    @property
    def spectrum_type(self) -> Optional[str]:
        return self._spectrum_type
    

    @property
    def start_background(self) -> Optional[str]:
        return self._start_background
    

    @property
    def end_background(self) -> Optional[str]:
        return self._end_background
    
    
    @property
    def background_spectrum(self) -> np.ndarray:
        if self._background_spectrum is None:
            self._background_spectrum = np.nanmean(self._dynamic_spectra[:, self._start_background_index:self._end_background_index+1], 
                                               axis=-1)
        return self._background_spectrum
    

    @property
    def dynamic_spectra_as_dBb(self) -> np.ndarray:
        if self._dynamic_spectra_as_dBb is None:
            # Create an artificial spectrogram where each spectrum is identically the background spectrum
            background_spectra = self.background_spectrum[:, np.newaxis]
            # Suppress divide by zero and invalid value warnings for this block of code
            with np.errstate(divide='ignore'):
                # Depending on the spectrum type, compute the dBb values differently
                if self._spectrum_type == SpectrumTypes.AMPLITUDE or self._spectrum_type == SpectrumTypes.DIGITS:
                    self._dynamic_spectra_as_dBb = 10 * np.log10(self._dynamic_spectra / background_spectra)
                elif self._spectrum_type == SpectrumTypes.POWER:
                    self._dynamic_spectra_as_dBb = 20 * np.log10(self._dynamic_spectra / background_spectra)
                else:
                    raise NotImplementedError(f"{self.spectrum_type} unrecognised, uncertain decibel conversion!")
        return self._dynamic_spectra_as_dBb  
    
    
    def set_background(self, 
                       start_background: str, 
                       end_background: str) -> None:
        """Public setter for start and end of the background"""
        self._dynamic_spectra_as_dBb = None # reset cache
        self._background_spectrum = None # reset cache
        self._start_background = start_background
        self._end_background = end_background
        self._update_background_indices_from_interval()
    
    
    
    def _update_background_indices_from_interval(self) -> None:
        start_background = datetime.strptime(self._start_background, TimeFormats.DATETIME)
        self._start_background_index = find_closest_index(start_background, 
                                                          self.datetimes, 
                                                          enforce_strict_bounds=True)
        
        end_background = datetime.strptime(self._end_background, TimeFormats.DATETIME)
        self._end_background_index = find_closest_index(end_background, 
                                                        self.datetimes, 
                                                        enforce_strict_bounds=True)


    def _check_shapes(self) -> None:
        num_spectrogram_dims = np.ndim(self._dynamic_spectra)
        # Check if 'dynamic_spectra' is a 2D array
        if num_spectrogram_dims != 2:
            raise ValueError(f"Expected dynamic spectrogram to be a 2D array, but got {num_spectrogram_dims}D array")
        dynamic_spectra_shape = self.dynamic_spectra.shape
        # Check if the dimensions of 'dynamic_spectra' are consistent with the time and frequency arrays
        if dynamic_spectra_shape[0] != self.num_frequencies:
            raise ValueError(f"Mismatch in number of frequency bins: Expected {self.num_frequencies}, but got {dynamic_spectra_shape[0]}")
        
        if dynamic_spectra_shape[1] != self.num_times:
            raise ValueError(f"Mismatch in number of time bins: Expected {self.num_times}, but got {dynamic_spectra_shape[1]}")
        

    def save(self) -> None:
        chunk_parent_path = get_chunks_dir_path(year  = self.chunk_start_datetime.year,
                                                month = self.chunk_start_datetime.month,
                                                day   = self.chunk_start_datetime.day)
        file_name = f"{self.chunk_start_time}_{self._tag}.fits"
        write_path = os.path.join(chunk_parent_path, 
                                  file_name)
        _save_spectrogram(write_path, self)
    

    def integrate_over_frequency(self, 
                                 correct_background: bool = False, 
                                 peak_normalise: bool = False) -> np.ndarray[np.float32]:
        
        # integrate over frequency
        I = np.trapz(self._dynamic_spectra, self._frequencies, axis=0)

        if correct_background:
            I = subtract_background(I, 
                                    self._start_background_index, 
                                    self._end_background_index)
        if peak_normalise:
            I = normalise_peak_intensity(I)
        return I


    def get_frequency_cut(self, 
                          at_time: float | str,
                          dBb: bool = False,
                          peak_normalise: bool = False) -> FrequencyCut:
        
        # it is important to note that the "at time" specified by the user likely does not correspond
        # exactly to one of the times assigned to each spectrogram. So, we compute the nearest achievable,
        # and return it from the function as output too.
        if isinstance(at_time, str):
            at_time = datetime.strptime(at_time, TimeFormats.DATETIME)
            index_of_cut = find_closest_index(at_time, 
                                              self.datetimes, 
                                              enforce_strict_bounds = True)
            time_of_cut = self.datetimes[index_of_cut]  

        elif isinstance(at_time, (float, int)):
            index_of_cut = find_closest_index(at_time, 
                                              self._times, 
                                              enforce_strict_bounds = True)
            time_of_cut = self.times[index_of_cut]
        
        else:
            raise ValueError(f"Type of at_time is unsupported: {type(at_time)}")
        
        if dBb:
            ds = self.dynamic_spectra_as_dBb
        else:
            ds = self._dynamic_spectra
        
        cut = ds[:, index_of_cut].copy() # make a copy so to preserve the spectrum on transformations of the cut

        if dBb:
            if peak_normalise:
                warn("Ignoring frequency cut normalisation, since dBb units have been specified")
        else:
            if peak_normalise:
                cut = normalise_peak_intensity(cut)
        
        return FrequencyCut(time_of_cut, 
                            self._frequencies, 
                            cut, 
                            self._spectrum_type)

        
    def get_time_cut(self,
                     at_frequency: float,
                     dBb: bool = False,
                     peak_normalise = False, 
                     correct_background = False, 
                     return_time_type: str = TimeTypes.SECONDS) -> TimeCut:
        
        # it is important to note that the "at frequency" specified by the user likely does not correspond
        # exactly to one of the physical frequencies assigned to each spectral component. So, we compute the nearest achievable,
        # and return it from the function as output too.
        index_of_cut = find_closest_index(at_frequency, 
                                          self._frequencies, 
                                          enforce_strict_bounds = True)
        frequency_of_cut = self.frequencies[index_of_cut]

        if return_time_type == TimeTypes.DATETIMES:
            times = self.datetimes
        elif return_time_type == TimeTypes.SECONDS:
            times = self.times
        else:
            raise ValueError(f"Invalid return_time_type. Got {return_time_type}, expected one of 'datetimes' or 'seconds'")

        # dependent on the requested cut type, we return the dynamic spectra in the preferred units
        if dBb:
            ds = self.dynamic_spectra_as_dBb
        else:
            ds = self.dynamic_spectra
        
        cut = ds[index_of_cut,:].copy() # make a copy so to preserve the spectrum on transformations of the cut

        # Warn if dBb is used with background correction or peak normalisation
        if dBb:
            if correct_background or peak_normalise:
                warn("Ignoring time cut normalisation, since dBb units have been specified")
        else:
            # Apply background correction if required
            if correct_background:
                cut = subtract_background(cut, 
                                          self._start_background_index,
                                          self._end_background_index)
            
            # Apply peak normalisation if required
            if peak_normalise:
                cut = normalise_peak_intensity(cut)

        return TimeCut(frequency_of_cut, 
                         times, 
                         cut,
                         self.spectrum_type)
    

def _seconds_of_day(dt: datetime) -> float:
    start_of_day = datetime(dt.year, dt.month, dt.day)
    return (dt - start_of_day).total_seconds()


# Function to create a FITS file with the specified structure
def _save_spectrogram(write_path: str, 
                      spectrogram: Spectrogram) -> None:
    
    capture_config = CaptureConfig(spectrogram.tag)
    ORIGIN    = capture_config.get_parameter_value(PNames.ORIGIN)
    INSTRUME  = capture_config.get_parameter_value(PNames.INSTRUMENT)
    TELESCOP  = capture_config.get_parameter_value(PNames.TELESCOPE)
    OBJECT    = capture_config.get_parameter_value(PNames.OBJECT)
    OBS_ALT   = capture_config.get_parameter_value(PNames.OBS_ALT)
    OBS_LAT   = capture_config.get_parameter_value(PNames.OBS_LAT)
    OBS_LON   = capture_config.get_parameter_value(PNames.OBS_LON)
    
    # Primary HDU with data
    primary_data = spectrogram.dynamic_spectra.astype(dtype=np.float32) 
    primary_hdu = fits.PrimaryHDU(primary_data)

    primary_hdu.header.set('SIMPLE', True, 'file does conform to FITS standard')
    primary_hdu.header.set('BITPIX', -32, 'number of bits per data pixel')
    primary_hdu.header.set('NAXIS', 2, 'number of data axes')
    primary_hdu.header.set('NAXIS1', spectrogram.dynamic_spectra.shape[1], 'length of data axis 1')
    primary_hdu.header.set('NAXIS2', spectrogram.dynamic_spectra.shape[0], 'length of data axis 2')
    primary_hdu.header.set('EXTEND', True, 'FITS dataset may contain extensions')

    # Add comments
    comments = [
        "FITS (Flexible Image Transport System) format defined in Astronomy and",
        "Astrophysics Supplement Series v44/p363, v44/p371, v73/p359, v73/p365.",
        "Contact the NASA Science Office of Standards and Technology for the",
        "FITS Definition document #100 and other FITS information."
    ]
    
    # The comments section remains unchanged since add_comment is the correct approach
    for comment in comments:
        primary_hdu.header.add_comment(comment)

    start_datetime = spectrogram.datetimes[0]
    start_date = start_datetime.strftime("%Y-%m-%d")
    start_time = start_datetime.strftime("%H:%M:%S.%f")

    end_datetime = spectrogram.datetimes[-1]
    end_date = end_datetime.strftime("%Y-%m-%d")
    end_time = end_datetime.strftime("%H:%M:%S.%f")

    primary_hdu.header.set('DATE', start_date, 'time of observation')
    primary_hdu.header.set('CONTENT', f'{start_date} dynamic spectrogram', 'title of image')
    primary_hdu.header.set('ORIGIN', f'{ORIGIN}')
    primary_hdu.header.set('TELESCOP', f'{TELESCOP}', 'type of instrument')
    primary_hdu.header.set('INSTRUME', f'{INSTRUME}') 
    primary_hdu.header.set('OBJECT', f'{OBJECT}', 'object description')

    primary_hdu.header.set('DATE-OBS', f'{start_date}', 'date observation starts')
    primary_hdu.header.set('TIME-OBS', f'{start_time}', 'time observation starts')
    primary_hdu.header.set('DATE-END', f'{end_date}', 'date observation ends')
    primary_hdu.header.set('TIME-END', f'{end_time}', 'time observation ends')

    primary_hdu.header.set('BZERO', 0, 'scaling offset')
    primary_hdu.header.set('BSCALE', 1, 'scaling factor')
    primary_hdu.header.set('BUNIT', f"{spectrogram.spectrum_type}", 'z-axis title') 

    primary_hdu.header.set('DATAMIN', np.nanmin(spectrogram.dynamic_spectra), 'minimum element in image')
    primary_hdu.header.set('DATAMAX', np.nanmax(spectrogram.dynamic_spectra), 'maximum element in image')

    primary_hdu.header.set('CRVAL1', f'{_seconds_of_day(start_datetime)}', 'value on axis 1 at reference pixel [sec of day]')
    primary_hdu.header.set('CRPIX1', 0, 'reference pixel of axis 1')
    primary_hdu.header.set('CTYPE1', 'TIME [UT]', 'title of axis 1')
    primary_hdu.header.set('CDELT1', spectrogram.time_resolution, 'step between first and second element in x-axis')

    primary_hdu.header.set('CRVAL2', 0, 'value on axis 2 at reference pixel')
    primary_hdu.header.set('CRPIX2', 0, 'reference pixel of axis 2')
    primary_hdu.header.set('CTYPE2', 'Frequency [Hz]', 'title of axis 2')
    primary_hdu.header.set('CDELT2', spectrogram.frequency_resolution, 'step between first and second element in axis')

    primary_hdu.header.set('OBS_LAT', f'{OBS_LAT}', 'observatory latitude in degree')
    primary_hdu.header.set('OBS_LAC', 'N', 'observatory latitude code {N,S}')
    primary_hdu.header.set('OBS_LON', f'{OBS_LON}', 'observatory longitude in degree')
    primary_hdu.header.set('OBS_LOC', 'W', 'observatory longitude code {E,W}')
    primary_hdu.header.set('OBS_ALT', f'{OBS_ALT}', 'observatory altitude in meter asl')


    # Wrap arrays in an additional dimension to mimic the e-CALLISTO storage
    times_wrapped = np.array([spectrogram.times.astype(np.float32)])
    # To mimic e-Callisto storage, convert frequencies to MHz
    frequencies_MHz = spectrogram.frequencies * 1e-6
    frequencies_wrapped = np.array([frequencies_MHz.astype(np.float32)])
    
    # Binary Table HDU (extension)
    col1 = fits.Column(name='TIME', format='PD', array=times_wrapped)
    col2 = fits.Column(name='FREQUENCY', format='PD', array=frequencies_wrapped)
    cols = fits.ColDefs([col1, col2])

    bin_table_hdu = fits.BinTableHDU.from_columns(cols)

    bin_table_hdu.header.set('PCOUNT', 0, 'size of special data area')
    bin_table_hdu.header.set('GCOUNT', 1, 'one data group (required keyword)')
    bin_table_hdu.header.set('TFIELDS', 2, 'number of fields in each row')
    bin_table_hdu.header.set('TTYPE1', 'TIME', 'label for field 1')
    bin_table_hdu.header.set('TFORM1', 'D', 'data format of field: 8-byte DOUBLE')
    bin_table_hdu.header.set('TTYPE2', 'FREQUENCY', 'label for field 2')
    bin_table_hdu.header.set('TFORM2', 'D', 'data format of field: 8-byte DOUBLE')
    bin_table_hdu.header.set('TSCAL1', 1, '')
    bin_table_hdu.header.set('TZERO1', 0, '')
    bin_table_hdu.header.set('TSCAL2', 1, '')
    bin_table_hdu.header.set('TZERO2', 0, '')

    # Create HDU list and write to file
    hdul = fits.HDUList([primary_hdu, bin_table_hdu])
    hdul.writeto(write_path, overwrite=True)



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Capture configuration files.
"""

from ._pvalidators  import PValidators
from ._capture_config import CaptureConfig
from ._parameters   import (
    Parameter, Parameters, parse_string_parameters, make_parameters
)
from ._capture_templates import (
    CaptureTemplate, CaptureModes, get_base_capture_template, make_base_capture_template
)
from ._pconstraints import (
    PConstraint, PConstraints, Bound, OneOf
)
from ._ptemplates   import (
    PTemplate, PNames, get_base_ptemplate, 
)

__all__ = [
    "Parameter", "Parameters", "parse_string_parameters", "make_parameters",
    "PConstraint", "PConstraints", "Bound", "OneOf", "PNames", "PTemplate",
    "get_base_ptemplate", "PValidators", "CaptureConfig", "CaptureTemplate",
    "CaptureModes",  "get_base_capture_template", "make_base_capture_template"
]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# SPDX-License-Identifier: GPL-3.0
#
# GNU Radio Python Flow Graph
# Title: Test receiver
# GNU Radio version: 3.10.1.1

# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

#
# RSPduo receiver top blocks
#

from functools import partial
from dataclasses import dataclass

from gnuradio import gr
from gnuradio import spectre
from gnuradio import sdrplay3

from spectre_core.capture_configs import Parameters, PNames
from spectre_core.config import get_chunks_dir_path
from ._base import capture


class _tuner_1_fixed_center_frequency(gr.top_block):
    def __init__(self, 
                 tag: str,
                 parameters: Parameters):
        gr.top_block.__init__(self, catch_exceptions=True)
        gr.top_block.__init__(self, "tuner_1_fixed", catch_exceptions=True)
        
        ##################################################
        # Unpack capture config
        ##################################################
        sample_rate = parameters.get_parameter_value(PNames.SAMPLE_RATE)
        batch_size  = parameters.get_parameter_value(PNames.BATCH_SIZE)
        center_freq = parameters.get_parameter_value(PNames.CENTER_FREQUENCY)
        bandwidth   = parameters.get_parameter_value(PNames.BANDWIDTH)
        if_gain     = parameters.get_parameter_value(PNames.IF_GAIN)
        rf_gain     = parameters.get_parameter_value(PNames.RF_GAIN)

        ##################################################
        # Blocks
        ##################################################
        self.spectre_batched_file_sink_0 = spectre.batched_file_sink(get_chunks_dir_path(), 
                                                                     tag, 
                                                                     batch_size, 
                                                                     sample_rate)
        self.sdrplay3_rspduo_0 = sdrplay3.rspduo(
            '',
            rspduo_mode="Single Tuner",
            antenna="Tuner 1 50 ohm",
            stream_args=sdrplay3.stream_args(
                output_type='fc32',
                channels_size=1
            ),
        )
        self.sdrplay3_rspduo_0.set_sample_rate(sample_rate)
        self.sdrplay3_rspduo_0.set_center_freq(center_freq)
        self.sdrplay3_rspduo_0.set_bandwidth(bandwidth)
        self.sdrplay3_rspduo_0.set_antenna("Tuner 1 50 ohm")
        self.sdrplay3_rspduo_0.set_gain_mode(False)
        self.sdrplay3_rspduo_0.set_gain(if_gain, 'IF')
        self.sdrplay3_rspduo_0.set_gain(rf_gain, 'RF')
        self.sdrplay3_rspduo_0.set_freq_corr(0)
        self.sdrplay3_rspduo_0.set_dc_offset_mode(False)
        self.sdrplay3_rspduo_0.set_iq_balance_mode(False)
        self.sdrplay3_rspduo_0.set_agc_setpoint(-30)
        self.sdrplay3_rspduo_0.set_rf_notch_filter(False)
        self.sdrplay3_rspduo_0.set_dab_notch_filter(False)
        self.sdrplay3_rspduo_0.set_am_notch_filter(False)
        self.sdrplay3_rspduo_0.set_biasT(False)
        self.sdrplay3_rspduo_0.set_debug_mode(False)
        self.sdrplay3_rspduo_0.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rspduo_0.set_show_gain_changes(False)


        ##################################################
        # Connections
        ##################################################
        self.connect((self.sdrplay3_rspduo_0, 0), (self.spectre_batched_file_sink_0, 0))
    
        
class _tuner_2_fixed_center_frequency(gr.top_block):
    def __init__(self, 
                 tag: str,
                 parameters: Parameters):
        gr.top_block.__init__(self, catch_exceptions=True)
        gr.top_block.__init__(self, "tuner_1_fixed", catch_exceptions=True)
        
        ##################################################
        # Unpack capture config
        ##################################################
        sample_rate = parameters.get_parameter_value(PNames.SAMPLE_RATE)
        batch_size  = parameters.get_parameter_value(PNames.BATCH_SIZE)
        center_freq = parameters.get_parameter_value(PNames.CENTER_FREQUENCY)
        bandwidth   = parameters.get_parameter_value(PNames.BANDWIDTH)
        if_gain     = parameters.get_parameter_value(PNames.IF_GAIN)
        rf_gain     = parameters.get_parameter_value(PNames.RF_GAIN)

        ##################################################
        # Blocks
        ##################################################
        self.spectre_batched_file_sink_0 = spectre.batched_file_sink(get_chunks_dir_path(), 
                                                                     tag, 
                                                                     batch_size, 
                                                                     sample_rate)
        self.sdrplay3_rspduo_0 = sdrplay3.rspduo(
            '',
            rspduo_mode="Single Tuner",
            antenna="Tuner 2 50 ohm",
            stream_args=sdrplay3.stream_args(
                output_type='fc32',
                channels_size=1
            ),
        )
        self.sdrplay3_rspduo_0.set_sample_rate(sample_rate)
        self.sdrplay3_rspduo_0.set_center_freq(center_freq)
        self.sdrplay3_rspduo_0.set_bandwidth(bandwidth)
        self.sdrplay3_rspduo_0.set_antenna("Tuner 2 50 ohm")
        self.sdrplay3_rspduo_0.set_gain_mode(False)
        self.sdrplay3_rspduo_0.set_gain(if_gain, 'IF')
        self.sdrplay3_rspduo_0.set_gain(rf_gain, 'RF', False)
        self.sdrplay3_rspduo_0.set_freq_corr(0)
        self.sdrplay3_rspduo_0.set_dc_offset_mode(False)
        self.sdrplay3_rspduo_0.set_iq_balance_mode(False)
        self.sdrplay3_rspduo_0.set_agc_setpoint(-30)
        self.sdrplay3_rspduo_0.set_rf_notch_filter(False)
        self.sdrplay3_rspduo_0.set_dab_notch_filter(False)
        self.sdrplay3_rspduo_0.set_am_notch_filter(False)
        self.sdrplay3_rspduo_0.set_biasT(False)
        self.sdrplay3_rspduo_0.set_stream_tags(False)
        self.sdrplay3_rspduo_0.set_debug_mode(False)
        self.sdrplay3_rspduo_0.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rspduo_0.set_show_gain_changes(False)


        ##################################################
        # Connections
        ##################################################
        self.connect((self.sdrplay3_rspduo_0, 0), (self.spectre_batched_file_sink_0, 0))    


class _tuner_1_swept_center_frequency(gr.top_block):
    def __init__(self, 
                 tag: str,
                 parameters: Parameters):
        gr.top_block.__init__(self, catch_exceptions=True)

        ##################################################
        # Unpack capture config
        ##################################################
        sample_rate      = parameters.get_parameter_value(PNames.SAMPLE_RATE)
        bandwidth        = parameters.get_parameter_value(PNames.BANDWIDTH)
        min_frequency    = parameters.get_parameter_value(PNames.MIN_FREQUENCY)
        max_frequency    = parameters.get_parameter_value(PNames.MAX_FREQUENCY)
        frequency_step   = parameters.get_parameter_value(PNames.FREQUENCY_STEP)
        samples_per_step = parameters.get_parameter_value(PNames.SAMPLES_PER_STEP)
        if_gain          = parameters.get_parameter_value(PNames.IF_GAIN)
        rf_gain          = parameters.get_parameter_value(PNames.RF_GAIN)
        batch_size       = parameters.get_parameter_value(PNames.BATCH_SIZE)

        ##################################################
        # Blocks
        ##################################################
        self.spectre_sweep_driver_0 = spectre.sweep_driver(min_frequency, 
                                                           max_frequency, 
                                                           frequency_step, 
                                                           sample_rate, 
                                                           samples_per_step,
                                                           'freq')
        self.spectre_batched_file_sink_0 = spectre.batched_file_sink(get_chunks_dir_path(), 
                                                                     tag, 
                                                                     batch_size, 
                                                                     sample_rate, 
                                                                     True, 
                                                                     'freq', 
                                                                     min_frequency)
        self.sdrplay3_rspduo_0 = sdrplay3.rspduo(
            '',
            rspduo_mode="Single Tuner",
            antenna="Tuner 1 50 ohm",
            stream_args=sdrplay3.stream_args(
                output_type='fc32',
                channels_size=1
            ),
        )
        self.sdrplay3_rspduo_0.set_sample_rate(sample_rate, True)
        self.sdrplay3_rspduo_0.set_center_freq(min_frequency, True)
        self.sdrplay3_rspduo_0.set_bandwidth(bandwidth)
        self.sdrplay3_rspduo_0.set_antenna("Tuner 1 50 ohm")
        self.sdrplay3_rspduo_0.set_gain_mode(False)
        self.sdrplay3_rspduo_0.set_gain(if_gain, 'IF', True)
        self.sdrplay3_rspduo_0.set_gain(rf_gain, 'RF', True)
        self.sdrplay3_rspduo_0.set_freq_corr(0)
        self.sdrplay3_rspduo_0.set_dc_offset_mode(False)
        self.sdrplay3_rspduo_0.set_iq_balance_mode(False)
        self.sdrplay3_rspduo_0.set_agc_setpoint(-30)
        self.sdrplay3_rspduo_0.set_rf_notch_filter(False)
        self.sdrplay3_rspduo_0.set_dab_notch_filter(True)
        self.sdrplay3_rspduo_0.set_am_notch_filter(False)
        self.sdrplay3_rspduo_0.set_biasT(False)
        self.sdrplay3_rspduo_0.set_stream_tags(True)
        self.sdrplay3_rspduo_0.set_debug_mode(False)
        self.sdrplay3_rspduo_0.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rspduo_0.set_show_gain_changes(False)


        ##################################################
        # Connections
        ##################################################
        self.msg_connect((self.spectre_sweep_driver_0, 'freq'), (self.sdrplay3_rspduo_0, 'freq'))
        self.connect((self.sdrplay3_rspduo_0, 0), (self.spectre_batched_file_sink_0, 0))
        self.connect((self.sdrplay3_rspduo_0, 0), (self.spectre_sweep_driver_0, 0))
        
        
@dataclass(frozen=True)
class CaptureMethods:
    tuner_1_fixed_center_frequency  = partial(capture, top_block_cls=_tuner_1_fixed_center_frequency)
    tuner_2_fixed_center_frequency  = partial(capture, top_block_cls=_tuner_2_fixed_center_frequency)
    tuner_1_swept_center_frequency  = partial(capture, top_block_cls=_tuner_1_swept_center_frequency)


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass

@dataclass
class PanelFormat:
    small_size      : int = 18
    medium_size     : int = 21
    large_size      : int = 24
    line_width      : int = 3
    style           : str = "dark_background"
    spectrogram_cmap: str = "gnuplot2"
    cuts_cmap       : str = "winter"
    integral_color  : str = "lime"

DEFAULT_FORMAT = PanelFormat()


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""
SPECTRE custom exceptions.
"""

class ChunkNotFoundError(FileNotFoundError): ...
class ChunkFileNotFoundError(FileNotFoundError): ...
class SpectrogramNotFoundError(FileNotFoundError): ...
class ModeNotFoundError(KeyError): ...
class EventHandlerNotFoundError(KeyError): ...
class ReceiverNotFoundError(KeyError): ...
class InvalidTagError(ValueError): ...
class InvalidSweepMetadataError(ValueError): ...



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime, timedelta
from typing import Tuple

import numpy as np
from astropy.io import fits
from astropy.io.fits.hdu.image import PrimaryHDU
from astropy.io.fits.hdu.table import BinTableHDU
from astropy.io.fits.hdu.hdulist import HDUList

from spectre_core.spectrograms import Spectrogram
from .._register import register_chunk
from .._base import BaseChunk, ChunkFile


@register_chunk('callisto')
class _Chunk(BaseChunk):
    def __init__(self, chunk_start_time: str, tag: str):
        super().__init__(chunk_start_time, tag) 
        self.add_file(FitsChunk(self.chunk_parent_path, self.chunk_name))


class FitsChunk(ChunkFile):
    def __init__(self, chunk_parent_path: str, chunk_name: str):
        super().__init__(chunk_parent_path, chunk_name, "fits")


    def read(self) -> Spectrogram:
        with fits.open(self.file_path, mode='readonly') as hdulist:
            primary_hdu = self._get_primary_hdu(hdulist)
            dynamic_spectra = self._get_dynamic_spectra(primary_hdu)
            microsecond_correction = self._get_microsecond_correction(primary_hdu)
            bintable_hdu = self._get_bintable_hdu(hdulist)
            times, frequencies = self._get_time_and_frequency(bintable_hdu)
            spectrum_type = self._get_spectrum_type(primary_hdu)

            if spectrum_type == "digits":
                dynamic_spectra_linearised = self._convert_units_to_linearised(dynamic_spectra)
                return Spectrogram(dynamic_spectra_linearised[::-1, :], # reverse the spectra along the frequency axis
                        times, 
                        frequencies[::-1], # sort the frequencies in ascending order
                        self.tag, 
                        chunk_start_time=self.chunk_start_time, 
                        microsecond_correction = microsecond_correction,
                        spectrum_type = spectrum_type)
            else:
                raise NotImplementedError(f"SPECTRE does not currently support spectrum type with BUNITS {spectrum_type}")


    @property
    def datetimes(self) -> np.ndarray:
        with fits.open(self.file_path, mode='readonly') as hdulist:
            bintable_data = hdulist[1].data
            times = bintable_data['TIME'][0]
            return [self.chunk_start_datetime + timedelta(seconds=t) for t in times]        


    def _get_primary_hdu(self, hdulist: HDUList) -> PrimaryHDU:
        return hdulist['PRIMARY']


    def _get_dynamic_spectra(self, primary_hdu: PrimaryHDU) -> np.ndarray:
        return primary_hdu.data


    def _get_microsecond_correction(self, primary_hdu: PrimaryHDU) -> int:
        date_obs = primary_hdu.header.get('DATE-OBS', None)
        time_obs = primary_hdu.header.get('TIME-OBS', None)
        datetime_obs = datetime.strptime(f"{date_obs}T{time_obs}", "%Y/%m/%dT%H:%M:%S.%f")
        return datetime_obs.microsecond


    def _get_bintable_hdu(self, hdulist: HDUList) -> BinTableHDU:
        return hdulist[1]


    def _get_time_and_frequency(self, bintable_hdu: BinTableHDU) -> Tuple[np.ndarray, np.ndarray]:
        data = bintable_hdu.data
        times = data['TIME'][0]
        frequencies_MHz = data['FREQUENCY'][0]
        frequencies = frequencies_MHz * 1e6 # convert to Hz
        return times, frequencies


    def _get_spectrum_type(self, primary_hdu: PrimaryHDU) -> str:
        return primary_hdu.header.get('BUNIT', None)


    def _convert_units_to_linearised(self, dynamic_spectra: np.ndarray) -> np.ndarray:
        digits_floats = np.array(dynamic_spectra, dtype='float')
        # conversion as per ADC specs [see email from C. Monstein]
        dB = (digits_floats / 255) * (2500 / 25)
        return 10 ** (dB / 10)





# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from ._decorators import log_call
from ._configure import configure_root_logger
from ._log_handlers import LogHandler, LogHandlers

__all__ = [
    "log_call", "configure_root_logger", "LogHandler", "LogHandlers"
]


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

from watchdog.observers import Observer
from watchdog.events import FileCreatedEvent

from ._factory import get_event_handler_from_tag
from spectre_core.config import get_chunks_dir_path

class PostProcessor:
    def __init__(self, 
                 tag: str):
        
        self._observer = Observer()

        EventHandler = get_event_handler_from_tag(tag)
        self._event_handler = EventHandler(tag)


    def start(self):
        """Start an observer to process newly created files in the chunks directory"""
        self._observer.schedule(self._event_handler, 
                                get_chunks_dir_path(), 
                                recursive=True,
                                event_filter=[FileCreatedEvent])
        
        try:
            _LOGGER.info("Starting the post processing thread...") 
            self._observer.start()
            self._observer.join()
        except KeyboardInterrupt:
            _LOGGER.warning(("Keyboard interrupt detected. Signalling "
                             "the post processing thread to stop"))
            self._observer.stop()
            _LOGGER.warning(("Post processing thread has been successfully stopped"))




# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from typing import Callable
from functools import wraps

def log_call(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args, **kwargs):
        logger = logging.getLogger(func.__module__)
        try:
            logger.info(f"Calling the function: {func.__name__}")
            return func(*args, **kwargs)
        except Exception as e:
            logger.error(f"Error in function: {func.__name__}", exc_info=True)
            raise
    return wrapper


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""
SPECTRE data paths.
"""

import os

_SPECTRE_DATA_DIR_PATH = os.environ.get("SPECTRE_DATA_DIR_PATH")
if _SPECTRE_DATA_DIR_PATH is None:
    raise ValueError("The environment variable SPECTRE_DATA_DIR_PATH has not been set")

_CHUNKS_DIR_PATH = os.environ.get("SPECTRE_CHUNKS_DIR_PATH", 
                                  os.path.join(_SPECTRE_DATA_DIR_PATH, 'chunks'))
os.makedirs(_CHUNKS_DIR_PATH, 
            exist_ok=True)

_LOGS_DIR_PATH = os.environ.get("SPECTRE_LOGS_DIR_PATH",
                               os.path.join(_SPECTRE_DATA_DIR_PATH, 'logs'))
os.makedirs(_LOGS_DIR_PATH, 
            exist_ok=True)

_CONFIGS_DIR_PATH = os.environ.get("SPECTRE_CONFIGS_DIR_PATH",
                                  os.path.join(_SPECTRE_DATA_DIR_PATH, "configs"))
os.makedirs(_CONFIGS_DIR_PATH, 
            exist_ok=True)


def get_spectre_data_dir_path(
) -> str:
    return _SPECTRE_DATA_DIR_PATH


def _get_date_based_dir_path(base_dir: str, year: int = None, 
                             month: int = None, day: int = None
) -> str:
    if day and not (year and month):
        raise ValueError("A day requires both a month and a year")
    if month and not year:
        raise ValueError("A month requires a year")
    
    date_dir_components = []
    if year:
        date_dir_components.append(f"{year:04}")
    if month:
        date_dir_components.append(f"{month:02}")
    if day:
        date_dir_components.append(f"{day:02}")
    
    return os.path.join(base_dir, *date_dir_components)


def get_chunks_dir_path(year: int = None, 
                        month: int = None, 
                        day: int = None
) -> str:
    return _get_date_based_dir_path(_CHUNKS_DIR_PATH, 
                                    year, 
                                    month, 
                                    day)


def get_logs_dir_path(year: int = None, 
                      month: int = None, 
                      day: int = None
) -> str:
    return _get_date_based_dir_path(_LOGS_DIR_PATH, 
                                    year, 
                                    month, 
                                    day)


def get_configs_dir_path(
) -> str:
    return _CONFIGS_DIR_PATH



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod
from typing import Callable, Optional
from numbers import Number

from spectre_core.capture_configs import (
    CaptureTemplate, CaptureModes, Parameters, Bound, PValidators, PNames,
    get_base_capture_template, get_base_ptemplate, OneOf, CaptureConfig
)
from .._base import BaseReceiver
from .._spec_names import SpecNames

class SDRPlayReceiver(BaseReceiver):
    def _get_pvalidator_fixed_center_frequency(self) -> Callable:
        def pvalidator(parameters: Parameters):
            PValidators.fixed_center_frequency(parameters)
        return pvalidator


    def _get_pvalidator_swept_center_frequency(self) -> None:
        def pvalidator(parameters: Parameters):
            PValidators.swept_center_frequency(parameters,
                                               self.get_spec(SpecNames.API_RETUNING_LATENCY))
        return pvalidator


    def _get_capture_template_fixed_center_frequency(self) -> CaptureTemplate:
        #
        # Create the base template
        #
        capture_template = get_base_capture_template( CaptureModes.FIXED_CENTER_FREQUENCY )
        capture_template.add_ptemplate( get_base_ptemplate(PNames.BANDWIDTH) )
        capture_template.add_ptemplate( get_base_ptemplate(PNames.IF_GAIN) )
        capture_template.add_ptemplate( get_base_ptemplate(PNames.RF_GAIN) )

        #
        # Update the defaults
        #
        capture_template.set_defaults(
            (PNames.BATCH_SIZE,            3.0),
            (PNames.CENTER_FREQUENCY,      95800000),
            (PNames.SAMPLE_RATE,           600000),
            (PNames.BANDWIDTH,             600000),
            (PNames.WINDOW_HOP,            512),
            (PNames.WINDOW_SIZE,           1024),
            (PNames.WINDOW_TYPE,           "blackman"),
            (PNames.RF_GAIN,               -30),
            (PNames.IF_GAIN,               -30)
        )   

        #
        # Adding pconstraints
        #
        capture_template.add_pconstraint(
            PNames.CENTER_FREQUENCY,
            [
                Bound(
                    lower_bound=self.get_spec(SpecNames.FREQUENCY_LOWER_BOUND),
                    upper_bound=self.get_spec(SpecNames.FREQUENCY_UPPER_BOUND)
                )
            ]
        )
        capture_template.add_pconstraint(
            PNames.SAMPLE_RATE,
            [
                Bound(
                    lower_bound=self.get_spec(SpecNames.SAMPLE_RATE_LOWER_BOUND),
                    upper_bound=self.get_spec(SpecNames.SAMPLE_RATE_UPPER_BOUND)
                )
            ]
        )
        capture_template.add_pconstraint(
            PNames.BANDWIDTH,
            [
                OneOf(
                    self.get_spec( SpecNames.BANDWIDTH_OPTIONS )
                )
            ]
        )
        capture_template.add_pconstraint(
            PNames.IF_GAIN,
            [
                Bound(
                    upper_bound=self.get_spec(SpecNames.IF_GAIN_UPPER_BOUND)
                )
            ]
        )
        capture_template.add_pconstraint(
            PNames.RF_GAIN,
            [
                Bound(
                    upper_bound=self.get_spec(SpecNames.RF_GAIN_UPPER_BOUND)
                )
            ]
        )
        return capture_template


    def _get_capture_template_swept_center_frequency(self) -> CaptureTemplate:
        #
        # Create the base template
        #
        capture_template = get_base_capture_template( CaptureModes.SWEPT_CENTER_FREQUENCY )
        capture_template.add_ptemplate( get_base_ptemplate(PNames.BANDWIDTH) )
        capture_template.add_ptemplate( get_base_ptemplate(PNames.IF_GAIN) )
        capture_template.add_ptemplate( get_base_ptemplate(PNames.RF_GAIN) )


        #
        # Update the defaults
        #
        capture_template.set_defaults(
            (PNames.BATCH_SIZE,            4.0),
            (PNames.MIN_FREQUENCY,         95000000),
            (PNames.MAX_FREQUENCY,         100000000),
            (PNames.SAMPLES_PER_STEP,      80000),
            (PNames.FREQUENCY_STEP,        1536000),
            (PNames.SAMPLE_RATE,           1536000),
            (PNames.BANDWIDTH,             1536000),
            (PNames.WINDOW_HOP,            512),
            (PNames.WINDOW_SIZE,           1024),
            (PNames.WINDOW_TYPE,           "blackman"),
            (PNames.RF_GAIN,               -30),
            (PNames.IF_GAIN,               -30)
        )   


        #
        # Adding pconstraints
        #
        capture_template.add_pconstraint(
            PNames.MIN_FREQUENCY,
            [
                Bound(
                    lower_bound=self.get_spec(SpecNames.FREQUENCY_LOWER_BOUND),
                    upper_bound=self.get_spec(SpecNames.FREQUENCY_UPPER_BOUND)
                )
            ]
        )
        capture_template.add_pconstraint(
            PNames.MAX_FREQUENCY,
            [
                Bound(
                    lower_bound=self.get_spec(SpecNames.FREQUENCY_LOWER_BOUND),
                    upper_bound=self.get_spec(SpecNames.FREQUENCY_UPPER_BOUND)
                )
            ]
        )
        capture_template.add_pconstraint(
            PNames.SAMPLE_RATE,
            [
                Bound(
                    lower_bound=self.get_spec(SpecNames.SAMPLE_RATE_LOWER_BOUND),
                    upper_bound=self.get_spec(SpecNames.SAMPLE_RATE_UPPER_BOUND)
                )
            ]
        )
        capture_template.add_pconstraint(
            PNames.BANDWIDTH,
            [
                OneOf(
                    self.get_spec( SpecNames.BANDWIDTH_OPTIONS )
                )
            ]
        )
        capture_template.add_pconstraint(
            PNames.IF_GAIN,
            [
                Bound(
                    upper_bound=self.get_spec(SpecNames.IF_GAIN_UPPER_BOUND)
                )
            ]
        )
        capture_template.add_pconstraint(
            PNames.RF_GAIN,
            [
                Bound(
                    upper_bound=self.get_spec(SpecNames.RF_GAIN_UPPER_BOUND)
                )
            ]
        )
        return capture_template